#include "BitVector.h"

#include <iostream> //For the pretty-printer
#include <stdlib.h> //for exit()
#include <sstream> //for to_string_dot()

#include <math.h>

#define INT_SIZE (sizeof(int)*8)


BitVector::BitVector(int size)
{
  this->size = size;
  bit_set.resize(ceil(((double)size)/INT_SIZE));
}


void 
BitVector::insert(int e)
{
  int mask = e / INT_SIZE;
  int offset = e % INT_SIZE;

  if(mask >= bit_set.size())
    bit_set.resize(mask+1);

  if(e >= size)
    size = e+1;

  bit_set[mask] = bit_set[mask] | (1 << offset);
}


void 
BitVector::insert(vector<int> v)
{
  for(int i=0; i<v.size(); i++)
    insert(v[i]);
}


int 
BitVector::has(int e)
{
  int mask = e / INT_SIZE;
  int offset = e % INT_SIZE;

  if(mask >= bit_set.size())
    return false;

  return ((bit_set[mask] & (1 << offset)) != 0);
}


int 
BitVector::operator ==(BitVector& b)
{
  if(size != b.size)
    return false;

  for(int i=0; i<bit_set.size(); i++)
    if(bit_set[i] != b.bit_set[i])
      return false;

  return true;
}


vector<int> 
BitVector::to_vector()
{
  vector<int> result;
  for(int i=0; i<size; i++)
    if(has(i))
      result.push_back(i);
  return result;
}



BitVector* 
BitVector::clone()
{
  BitVector* result = new BitVector(size);
  for(int i=0; i<bit_set.size(); i++)
    result->bit_set[i] = bit_set[i];
  return result;
}


BitVector& operator |(BitVector& b1, BitVector& b2)
{
  vector<int> v1 = b1.bit_set;
  vector<int> v2 = b2.bit_set;

  int max_size;
  if(v1.size() > v2.size())
    v2.resize(v1.size());
  else
    v1.resize(v2.size());

  if(b1.size>b2.size)
    max_size = b1.size;
  else
    max_size = b2.size;
  //size = max(size(b1),size(b2))

  BitVector* result = new BitVector(max_size);
  for(int i=0; i<v1.size(); i++)
    result->bit_set[i] = v1[i] | v2[i];
  //union

  return *result;
}


BitVector& operator &(BitVector& b1, BitVector& b2)
{
  vector<int> v1 = b1.bit_set;
  vector<int> v2 = b2.bit_set;

  int min_size;
  int min_mask;
  if(v1.size() > v2.size())
    min_mask = v2.size();
  else
    min_mask = v1.size();

  if(b2.size>b1.size)
    min_size = b1.size;
  else
    min_size = b2.size;
  //size = min(size(b1),size(b2))

  BitVector* result = new BitVector(min_size);
  for(int i=0; i<min_mask; i++)
    result->bit_set[i] = v1[i] & v2[i];
  //intersection

  return *result;
}


BitVector& operator -(BitVector& b1, BitVector& b2)
{
  vector<int> v1 = b1.bit_set;
  vector<int> v2 = b2.bit_set;

  int max_size;
  if(v1.size() > v2.size())
    v2.resize(v1.size());
  else
    v1.resize(v2.size());

  if(b1.size>b2.size)
    max_size = b1.size;
  else
    max_size = b2.size;
  //size = max(size(b1),size(b2))

  BitVector* result = new BitVector(max_size);
  for(int i=0; i<v1.size(); i++)
    result->bit_set[i] = v1[i] & (~(v2[i]));
  //A - B = A n (complement(B))

  return *result;
}


string 
BitVector::to_string_dot()
{
  stringstream sout;

  sout << "{";
  for(int i=0; i<size; i++)
    if(has(i))
      sout << "r" << i << " ";
  sout << "}";

  return sout.str();
}


ostream& operator <<(ostream& sout, BitVector& b)
{
  sout << "{";
  for(int i=0; i<b.size; i++)
    if(b.has(i))
      sout << "r" << i << " ";
  sout << "}";
  // [size=" << b.size << ", masks=" << b.bit_set.size() << "]";
  return sout;
}


