#ifndef __CFG_H
#define __CFG_H

#include "BitVector.h"
#include "PseudoCode.h"

#include <iostream> //For the pretty-printer

#include <vector>
#include <map>

using namespace std;

class Cfg
{
 public:
  vector<Code*> node;
  
  vector<vector<int> > succ; 
  //succ[i] = successors for node i

  map<string,int> label_node;
  //label_node["loop"] = #node 'loop:'

  vector<BitVector*> live_in;
  vector<BitVector*> live_out;

  Cfg();

  void add_node(Code* node);
  void set_live_in(BitVector* node_live_in);   //
  void set_live_out(BitVector* node_live_out); // for last node

  void fix_succ();

  //
  // Liveness analysis
  //
  void do_liveness(); //compute live-in and live-out sets

  //
  // Basic blocks extraction
  //

  Cfg* extract_basic_blocks();


  //Pretty-printers
  void print_dot(ostream& sout);
  void show_dot();
  void print_code(); //print code only

  Cfg* clone();
};

//Cfg pretty constructors (fill a global cfg)
void add(int r_dest, int r1, int r2);
void sub(int r_dest, int r1, int r2);
void mul(int r_dest, int r1, int r2);
void div(int r_dest, int r1, int r2);
void mov_reg_int(int r_dest, int constante);
void copy(int r_dest, int r_source);
void load(int r_dest, int r_base, int shift);
void store(int r_dest, int shift, int r_source);
void jump(string label);
void cjump(int r1, test_t relation, int r2, string label);
void call(string label);
void ret(int nb_args);
void label(string label);

void push_all();
void pop_all();
void push(int reg);
void pop(int reg);
void load_arp(int reg, int shift);
void store_arp(int shift, int reg);
void add_sp_reg(int reg);
void sub_sp_reg(int reg);
void mov_reg_arp(int reg);
void mov_reg_sp(int reg);
void mov_sp_reg(int reg);
void push_arp();
void pop_arp();
void mov_arp_sp();
void mov_sp_arp();

//Pretty printer
ostream& operator <<(ostream& sout, Cfg& p);


#endif
